
/*
 * faderoot.c 	utility to s_l_o_w_l_y change the color of the
 *      X Window System root window
 *
 *  Author:	Sam Denton, St. Louis, MO
 *		13-Sep-93  (Happy Birthday, Dad!)
 *
 * Copyright 1993 Samuel T. Denton, III
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that my name not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior
 * permission.  I make no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 * 
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/bitmaps/gray>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>

#define Dynamic 1

/*
 * srand/rand isn't a very good random number
 * generator, but hey, this isn't brain surgery.
 */

#define RANDOM_INIT srand(getpid())
#define RANDOM_LARGE (0xFFFF & rand())
#define RANDOM_SMALL (0x80 + (0x7F & rand()))

/* globals */

char *program_name;
Display *dpy;
int screen;
Window root;
Pixmap save_pixmap = (Pixmap)None;
XColor color;

usage()
{
    fprintf(stderr, "usage: %s [options]\n", program_name);
    fprintf(stderr, "  where options are:\n");
    fprintf(stderr, "  -display <display>\n");
    fprintf(stderr, "  -time <interval>\n");
    fprintf(stderr, "  -help\n");
    fprintf(stderr, "  -solid\n");
    fprintf(stderr, "  -gray   or   -grey\n");
    fprintf(stderr, "  -bitmap <filename>\n");
    fprintf(stderr, "  -mod <x> <y>\n");
    exit(1);
    /*NOTREACHED*/
}

Pixmap MakeModuloBitmap(), ReadBitmapFile();
void BounceComponent(), Daemonize();

main(argc, argv) 
    int argc;
    char **argv;
{
    int excl = 0;
    char *display_name = NULL;
    int solid_color = 0;
    int gray = 0;
    char *bitmap_file = NULL;
    int mod_x = 0;
    int mod_y = 0;
    register int i;
    unsigned int ww, hh;
    Pixmap bitmap;
    int interval = 1;
    XWindowAttributes attr;
    unsigned long mask;
    int dr, dg, db;

    program_name=argv[0];

    for (i = 1; i < argc; i++) {
	if (!strcmp ("-display", argv[i]) || !strcmp ("-d", argv[i])) {
	    if (++i>=argc) usage ();
	    display_name = argv[i];
	    continue;
	}
	if (!strcmp("-time", argv[i])) {
	    if (++i>=argc) usage();
	    interval = atoi(argv[i]);
	    if (interval < 0) interval = 1;
	    continue;
	}
	if (!strcmp("-help", argv[i])) {
	    usage();
	}
	if (!strcmp("-solid", argv[i])) {
	    solid_color = 1;
	    excl++;
	    continue;
	}
	if (!strcmp("-gray", argv[i]) || !strcmp("-grey", argv[i])) {
	    gray = 1;
	    excl++;
	    continue;
	}
	if (!strcmp("-bitmap", argv[i])) {
	    if (++i>=argc) usage();
	    bitmap_file = argv[i];
	    excl++;
	    continue;
	}
	if (!strcmp("-mod", argv[i])) {
	    if (++i>=argc) usage();
	    mod_x = atoi(argv[i]);
	    if (mod_x <= 0) mod_x = 1;
	    if (++i>=argc) usage();
	    mod_y = atoi(argv[i]);
	    if (mod_y <= 0) mod_y = 1;
	    excl++;
	    continue;
	}
	usage();
    } 

    /* Check for no or multiple use of exclusive options */
    if (excl > 1) {
	fprintf(stderr, "%s: choose only one of {solid, gray, bitmap, mod}\n",
		program_name);
	usage();
    }

    /* If there are no arguments then complain. */
    if (!excl) {
	fprintf(stderr, "%s: choose one of {solid, gray, bitmap, mod}\n",
		program_name);
	usage();
    }

    dpy = XOpenDisplay(display_name);
    if (!dpy) {
	fprintf(stderr, "%s:  unable to open display '%s'\n",
		program_name, XDisplayName (display_name));
	exit (2);
    }
    screen = DefaultScreen(dpy);
    if (!(DefaultVisual(dpy, screen)->class & Dynamic)) {    
	fprintf(stderr, "%s:  default visual for '%s' is static\n",
		program_name, XDisplayName (display_name));
	exit (3);
    }
    root = RootWindow(dpy, screen);
    XGetWindowAttributes(dpy, root, &attr);
    XAllocColorCells(dpy, attr.colormap, False, &mask, 0, &color.pixel, 1);

    /* Handle -solid option */
    if (solid_color) {
	XSetWindowBackground(dpy, root, color.pixel);
    }

    /* Handle -bitmap option */
    if (bitmap_file) {
	bitmap = ReadBitmapFile(bitmap_file, &ww, &hh,
			(int *)NULL, (int *)NULL);
	SetBackgroundToBitmap(bitmap, ww, hh);
    }

    /* Handle set background to a modulo pattern */
    if (mod_x) {
	bitmap = MakeModuloBitmap(mod_x, mod_y);
	SetBackgroundToBitmap(bitmap, 16, 16);
    }

    /* Handle -gray and -grey options */
    if (gray) {
	bitmap = XCreateBitmapFromData(dpy, root, (const char*)gray_bits,
				       gray_width, gray_height);
	SetBackgroundToBitmap(bitmap, gray_width, gray_height);
    }

    XClearWindow(dpy, root);
    FixupState();
    if (interval) Daemonize();

    /* randomize everything */
    RANDOM_INIT;
    color.red = RANDOM_LARGE;
    color.green = RANDOM_LARGE;
    color.blue = RANDOM_LARGE;
    color.flags = DoRed | DoGreen | DoBlue;
    dr = RANDOM_SMALL;
    dg = RANDOM_SMALL;
    db = RANDOM_SMALL;

    /* do the work */
    for (;;) {
	if (!XStoreColor(dpy, attr.colormap, &color)) break;
	if (!XFlush(dpy)) break;
	if (interval) sleep(interval);
	BounceComponent(&color.red, &dr);
	BounceComponent(&color.green, &dg);
	BounceComponent(&color.blue, &db);
    }
    fprintf(stderr, "%s:  connection broken by server\n", program_name);
    exit (0);
}

/* Free past incarnation if needed, and retain state if needed. */
FixupState()
{
    Atom prop, type;
    int format;
    unsigned long length, after;
    unsigned char *data;

    prop = XInternAtom(dpy, "_XSETROOT_ID", False);
    if (DefaultVisual(dpy, screen)->class & Dynamic) {    
	(void)XGetWindowProperty(dpy, root, prop, 0L, 1L, True, AnyPropertyType,
				 &type, &format, &length, &after, &data);
	if ((type == XA_PIXMAP) && (format == 32) &&
	    (length == 1) && (after == 0))
	    XKillClient(dpy, *((Pixmap *)data));
	else if (type != None)
	    fprintf(stderr, "%s: warning: _XSETROOT_ID property is garbage\n",
		    program_name);
    }
    if (!save_pixmap)
	save_pixmap = XCreatePixmap(dpy, root, 1, 1, 1);
    XChangeProperty(dpy, root, prop, XA_PIXMAP, 32, PropModeReplace,
			(unsigned char *) &save_pixmap, 1);
    XSetCloseDownMode(dpy, RetainPermanent);
}

/*
 * SetBackgroundToBitmap: Set the root window background to a caller supplied 
 *                        bitmap.
 */
SetBackgroundToBitmap(bitmap, width, height)
    Pixmap bitmap;
    unsigned int width, height;
{
    GC gc;
    XGCValues gc_init;

    gc_init.background = color.pixel;
    gc_init.foreground = BlackPixel(dpy, screen);
    gc = XCreateGC(dpy, root, GCForeground|GCBackground, &gc_init);
    save_pixmap = XCreatePixmap(dpy, root, width, height,
			(unsigned int)DefaultDepth(dpy, screen));
    XCopyPlane(dpy, bitmap, save_pixmap, gc, 0, 0, width, height,
			0, 0, (unsigned long)1);
    XSetWindowBackgroundPixmap(dpy, root, save_pixmap);
    XFreeGC(dpy, gc);
    XFreePixmap(dpy, bitmap);
}


/*
 * MakeModuloBitmap: Returns a modulo bitmap based on an x & y mod.
 */
Pixmap
MakeModuloBitmap(mod_x, mod_y)
    int mod_x, mod_y;
{
    int i;
    long pattern_line = 0;
    char modulo_data[16*16/8];

    for (i=0; i<16; i++) {
	pattern_line <<=1;
	if ((i % mod_x) == 0) pattern_line |= 0x0001;
    }
    for (i=0; i<16; i++) {
	if ((i % mod_y) == 0) {
	    modulo_data[i*2] = 0xff;
	    modulo_data[i*2+1] = 0xff;
	} else {
	    modulo_data[i*2] = pattern_line & 0xff;
	    modulo_data[i*2+1] = (pattern_line>>8) & 0xff;
	}
    }

    return(XCreateBitmapFromData(dpy, root, modulo_data, 16, 16));
}

Pixmap
ReadBitmapFile(filename, width, height, x_hot, y_hot)
    char *filename;
    unsigned int *width, *height;
    int *x_hot, *y_hot;
{
    Pixmap bitmap;
    int status;

    status = XReadBitmapFile(dpy, root, filename, width,
			     height, &bitmap, x_hot, y_hot);
    if (status == BitmapSuccess)
      return(bitmap);
    else if (status == BitmapOpenFailed)
	fprintf(stderr, "%s: can't open file: %s\n", program_name, filename);
    else if (status == BitmapFileInvalid)
	fprintf(stderr, "%s: bad bitmap format file: %s\n",
			program_name, filename);
    else
	fprintf(stderr, "%s: insufficient memory for bitmap: %s",
			program_name, filename);
    exit(1);
    /*NOTREACHED*/
}

void
BounceComponent(value, delta)
    unsigned short *value;
    int *delta;
{
    long tmp = (long)*value + (long)*delta;
    if (tmp<0) {
	tmp = 0;
	*delta = RANDOM_SMALL;
    } else if (tmp>0xFFFF) {
	tmp = 0xFFFF;
	*delta = -(RANDOM_SMALL);
    }
    *value = tmp;
}

void
Daemonize()
{
    int pid, tty;

    /* clone ourself */
    pid = fork();
    if ( pid < 0 ) {
	perror( "fork" );
	exit( 1 );
    } else if ( pid > 0 ) {
	/* Parent just exits. */
	exit( 0 );
    }

    /* Go stealth (ditch our controlling tty) */
	for (tty=0; tty<3; tty++) close(tty);
#ifndef __svr4__
    setpgrp(0, getpid());
    if ((tty = open("/dev/tty", 0) >= 0)) {
	ioctl(tty, TIOCNOTTY, 0);
	(void) close( tty );
    }
#else /* __svr4__ */
    setpgrp();
#endif /* __svr4__ */
}
